/*======================================================================
:::
::: ColorLib.cpp
::: 
::: Platform-independent color manipulation routines for 
::: the TealPoint image manager
:::
::: by TealPoint Software (c)1997-2003, All Rights Reserved
:::
======================================================================*/

/*==========================================
// Includes
//========================================*/

	#include <stdio.h>
	#include <stdarg.h>
	#include <string.h>
#if macintosh
	#include <stdlib.h>
#else
	#include <malloc.h>
#endif

	#include <ctype.h>
	#include <math.h>

#define ABS(a) 	((a)>0?(a):-(a))

/*============================================================
:::
::: color_RGB2HLS()
:::
::: Convert an RGB color to HLS
:::
============================================================*/

void color_RGB2HLS( short r, short g, short b, short *h, short *l, short *s )
{
	int maxval,minval;
	int diff;
	int dr,dg,db;

	maxval = (r>g && r>b) ? r : (g>b) ? g : b;
	minval = (r<g && r<b) ? r : (g<b) ? g : b;

	diff = maxval-minval;

	*l   = (maxval+minval)/2;

	if (!diff) {
		*s = 0;
		*h = 0;
	} else {
		if (*l <= 127) *s = 255 * diff/(    maxval+minval);
		else           *s = 255 * diff/(510-maxval-minval);

		dr = 255 * (maxval-r)/diff;
		dg = 255 * (maxval-g)/diff;
		db = 255 * (maxval-b)/diff;

		if      (r==maxval) *h = db - dg;
		else if (g==maxval) *h = 510 + dr - db;
		else if (b==maxval) *h = 1020 + dg - dr;

		*h /= 6;

		if (*h<0) *h+=255;
	}
}

/*============================================================
:::
::: color_HLS2RGB()
:::
::: Convert an HLS color to RGB
:::
============================================================*/

	static long rgb( long q1, long q2, int hue )
	{
		if (hue>255) hue-= 255;
		if (hue<0  ) hue+= 255;
		if (hue < (255/6)) return q1 + (q2-q1) * hue * 6/255;
		else if (hue < (255/2)) return q2;
			else if (hue < (170)) return q1 + (q2-q1) * (170-hue) * 6/255;
				else return q1;
	}

void color_HLS2RGB( short h, short l, short s, short *r, short *g, short *b )
{
	long p1,p2;

	if (l <= 127) p2 = l * (s+255);   // P2 is 255-times'd
	else p2 = 255*(l+s) - l*s;

	p1 = 510*l - p2;

	if (!s) {
		*r = *g = *b = l;
	} else {
		*r = rgb( p1,p2, h+85 )/255;
		*g = rgb( p1,p2, h    )/255;
		*b = rgb( p1,p2, h-85 )/255;

		if (*r>255) *r = 255;
		if (*g>255) *g = 255;
		if (*b>255) *b = 255;
	}
}


/*============================================================
:::
::: color_AdjustBrightnessContrast()
:::
============================================================*/

short color_AdjustBrightnessContrast( short *r, short *g, short *b, short brightness, short contrast )
{
	short h,l,s;
	long base  = (100+brightness)*128/100;
	long scale = 65536L * (contrast + 100)/100;

	color_RGB2HLS( *r,*g,*b, &h,&l,&s );

	l = (short)(base + (((l-128)*scale)>>16));
	if (l<0) l=0;
	else if (l>255) l=255;

	color_HLS2RGB( h,l,s, r,g,b );
	return 0;
}


/*============================================================
:::
::: color_ConvertPixel()
:::
============================================================*/

	#define	DIV6	(255/6+1)
	#define	DIV4	(255/4+1)
	#define	DIV8	(255/8+1)
	#define	DIV16	(255/16+1)

//	#define	MUL6	(255/5)
//	#define	MUL4	(255/3)
//	#define	MUL16	(255/15)

	#define	MUL6	DIV6
	#define	MUL4	DIV4
	#define	MUL16	DIV16

	static unsigned char dither_pattern[8][8] = {

		{	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff 	}, //	8/8
		{	0xbb,0xff,0xee,0xff,0xbb,0xff,0xee,0xff 	}, //	7/8
		{	0xaa,0xff,0xaa,0xff,0xaa,0xff,0xaa,0xff 	}, //	6/8
		{	0xaa,0x77,0xaa,0xdd,0xaa,0x77,0xaa,0xdd 	}, //	5/8
		{	0xaa,0x55,0xaa,0x55,0xaa,0x55,0xaa,0x55 	}, //	4/8
//		{	0xaa,0x11,0xaa,0x44,0xaa,0x11,0xaa,0x44		}, //	3/8
		{	0xaa,0x00,0xaa,0x00,0xaa,0x00,0xaa,0x00 	}, //	2/8
		{	0x88,0x00,0x22,0x00,0x88,0x00,0x22,0x00 	}, //	1/8
		{	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00		}  //;1/8
	};

short color_ConvertPixel( short *r, short *g, short *b, short depth, short dither, short x, short y )
{
	if (dither) {
		short sum = (*r+*g+*g+*b)>>2,val;
		switch( depth ) {
			case 1:	
				*r = *g = *b = ((~dither_pattern[sum/DIV8][y&7] & (128>>(x&7))) ? 255:0); 
				break;
					
			case 2:
				sum += DIV4/2;
				val = sum / DIV4;	
				if (val && !(~dither_pattern[ (sum % DIV4) * 8 / DIV4 ][y&7] & (128>>(x&7)) )) val--;
				if (val >= 4) val--;
				*r = *g = *b = val * MUL4;
				break;

			case 4:	
				sum += DIV16/2;
				val = sum / DIV16;	
				if (val && !(~dither_pattern[ (sum % DIV16) * 8 / DIV16 ][y&7] & (128>>(x&7)) )) val--;
				if (val >= 16) val--;
				*r = *g = *b = val * MUL16;
				break;

			case 8:	
				if ( ((*r)&240)==((*g)&240) && ((*g)&240)==((*b)&240) ) {
					val = (*g+16)/17*17;
					if (val && !(~dither_pattern[ *g % 17 * 8 / 17][y&7] & (128>>(x&7)) )) val--;
					*r = *g = *b = val;
				} else {
					*r += DIV6/2;
					val = *r / DIV6;	
					if (val && !(~dither_pattern[ (*r % DIV6) * 8 / DIV6 ][y&7] & (128>>(x&7)) )) val--;
					if (val >= 6) val--;
					*r = val * MUL6;

					*g += DIV6/2;
					val = *g / DIV6;	
					if (val && !(~dither_pattern[ (*g % DIV6) * 8 / DIV6 ][y&7] & (128>>(x&7)) )) val--;
					if (val >= 6) val--;
					*g = val * MUL6;

					*b += DIV6/2;
					val = *b / DIV6;	
					if (val && !(~dither_pattern[ (*b % DIV6) * 8 / DIV6 ][y&7] & (128>>(x&7)) )) val--;
					if (val >= 6) val--;
					*b = val * MUL6;
				}
				break;

			case 16:
				val = *r & 0xf8; 
				if (val<0xf8 && (~dither_pattern[  *r & 7     ][y&7] & (128>>(x&7)) )) val += 0x08;
				*r = val;

				val = *g & 0xfc; 
				if (val<0xfc && (~dither_pattern[ (*g & 3)<<1 ][y&7] & (128>>(x&7)) )) val += 0x04;
				*g = val;

				val = *b & 0xf8; 
				if (val<0xf8 && (~dither_pattern[  *b & 7     ][y&7] & (128>>(x&7)) )) val += 0x08;
				*b = val;
				break;
		}

	} else {
		switch( depth ) {
			case 1:	
				*r = *g = *b = ((*r+*g+*b)/3 > 127 ? 255:0); 
				break;

			case 2:	
				*r = *g = *b = (*r+*g+*b)/3 / (256/3) * (255/3);	
				break;

			case 4:	
				*r = *g = *b = (*r+*g+*b)/3 / (256/7) * (255/7);	
				break;

			case 8:	
				if ( ((*r)&240)==((*g)&240) && ((*g)&240)==((*b)&240) ) {
					*r = *g = *b = *g - (*g%17);
				} else {
					*r = *r/DIV6 * MUL6; 
					*g = *g/DIV6 * MUL6; 
					*b = *b/DIV6 * MUL6;
				}
				break;

			case 16:	
				*r &= 0xf8; 
				*g &= 0xfc; 
				*b &= 0xf8; 
				break;
		}
	}
	return 0;
}


/*============================================================
:::
::: color_GetScaledBitmapRGB()
:::
::: Scale an RGB bitmap buffer to the target size
:::
============================================================*/

short color_GetScaledBitmapRGB( char **tbuf, short twidth, short theight, char *sbuf, short swidth, short sheight )
{
	long i,j,m,n,delt;
	long isx,isy,iex,iey;
	long fwidth,fheight;
	float sx,sy,ex,ey,fx,fy;
	float r,g,b,t;
	unsigned char *buffer, *fptr, *tptr;

	fwidth  = twidth;
	fheight = theight;

	//============================================== 
	// Allocate target buffer
	//==============================================

	if (buffer = (unsigned char *)malloc( (long)fwidth*fheight*3 )) {
		tptr = buffer;

		memset( buffer, 255, (long)fwidth*fheight*3 );

		//
		// Go Through Target Pixels
		//

		for (j=0; j<theight; j++) {
			for (i=0; i<twidth; i++) {

				sx = (float) i   *(float)swidth /(float)twidth;
				ex = (float)(i+1)*(float)swidth /(float)twidth;
				sy = (float) j   *(float)sheight/(float)theight;
				ey = (float)(j+1)*(float)sheight/(float)theight;

				isx = (int)floor(sx);
				iex = (int)ceil (ex);
				isy = (int)floor(sy);
				iey = (int)ceil (ey);

				r = g = b = t = (float)0.0;

				delt = swidth*isy+isx;
				fptr = (unsigned char *)sbuf + delt+delt+delt;

				//
				// Go Through Source Pixels
				//

				for (n=isy; n<iey; n++) {

					fy = ((float)n < sy) ? (sy-(float)n) : ((float)n > ey) ? ((float)n-ey) : (float)1.0;

					for (m=isx; m<iex; m++) {

						fx = ((float)m < sx) ? (sx-(float)m) : ((float)m > ex) ? ((float)m-ex) : (float)1.0;
						fx *= fy;

						r += (float)fptr[2] * fx;
						g += (float)fptr[1] * fx;
						b += (float)fptr[0] * fx;
						t += fx;

						fptr += 3;
					}
					delt  = swidth - (iex-isx);
					fptr += delt+delt+delt;
				}

				if (t) {
					t = (float)1.0/t;
					*tptr++ = (unsigned char)(b*t);
					*tptr++ = (unsigned char)(g*t);
					*tptr++ = (unsigned char)(r*t);
				} else {
					delt = swidth*isy+isx;
					fptr = (unsigned char *)sbuf + delt+delt+delt;
					*tptr++ = *fptr++;
					*tptr++ = *fptr++;
					*tptr++ = *fptr++;
				}
			}
			delt = fwidth-twidth;
			tptr += delt+delt+delt;
		}

		*tbuf = (char *)buffer;		
		return 0;
	}

	free(buffer);
	return -1;
}


/*============================================================
:::
::: color_IndexInit()
:::
============================================================*/

	#define COLOR_DITHER 3

	static int color2index[256];
	static int gray2index [256];

	static int gray_scale [] = { 34,68,85,119,136,170,187,221 };
	static int color_scale[] = { 17,51,102,153,204,255 };

	static int r_off[] = { 0, 18, 36, 54, 72, 90 };
	static int g_off[] = { 0,  1,  2,  3,  4,  5 };
	static int b_off[] = { 0,  6, 12,108,114,120 };

void color_IndexInit( void )
{
	int i;

	for (i=0;i<256;i++) {
		if (i<(color_scale[0]+color_scale[1])/2) color2index[i] = 0;
		else if (i<(color_scale[1]+color_scale[2])/2) color2index[i] = 1;
		else if (i<(color_scale[2]+color_scale[3])/2) color2index[i] = 2;
		else if (i<(color_scale[3]+color_scale[4])/2) color2index[i] = 3;
		else if (i<(color_scale[4]+color_scale[5])/2) color2index[i] = 4;
		else color2index[i] = 5;
	}

	for (i=0;i<256;i++) {
		if (i<(gray_scale[0]+gray_scale[1])/2) gray2index[i] = 0;
		else if (i<(gray_scale[1]+gray_scale[2])/2) gray2index[i] = 1;
		else if (i<(gray_scale[2]+gray_scale[3])/2) gray2index[i] = 2;
		else if (i<(gray_scale[3]+gray_scale[4])/2) gray2index[i] = 3;
		else if (i<(gray_scale[4]+gray_scale[5])/2) gray2index[i] = 4;
		else if (i<(gray_scale[5]+gray_scale[6])/2) gray2index[i] = 5;
		else if (i<(gray_scale[6]+gray_scale[7])/2) gray2index[i] = 6;
		else gray2index[i] = 7;
	}
}

/*============================================================
:::
::: color_IndexFromRGB()
:::
============================================================*/

short color_IndexFromRGB( short i, short k, short r, short g, short b )
{
	int val;

	// Add dither

	if ((i+k)&1) {
		r += COLOR_DITHER;
		g += COLOR_DITHER;
		b += COLOR_DITHER;
		if (r>255) r=255;
		if (g>255) g=255;
		if (b>255) b=255;
	} else {
		r -= COLOR_DITHER;
		g -= COLOR_DITHER;
		b -= COLOR_DITHER;
		if (r<0) r=0;
		if (g<0) g=0;
		if (b<0) b=0;
	}

	// Black?

	if (r<8 && g<8 && b<8) {
		val = 255;
	} 

	// Color?

	else {
		val = (r+g+b)/3;

		r = color2index[r];
		g = color2index[g];
		b = color2index[b];

		// greyscale?

		if (r==g && g==b && 
			ABS((val- gray_scale[ gray2index[val]])) <
			ABS((val-color_scale[color2index[val]])) ) {
			val = 6*6*6+gray2index[val];
		} else {

			val = r_off[5-r] + g_off[5-g] + b_off[5-b];
		}
	}

	return val;
}


/*============================================================
:::
::: color_ReduceRgbBitmap()
:::
::: Convert a 24-bit bitmap to 1-16 bits depth (palm format)
:::
============================================================*/

short color_ReduceRgbBitmap( char *buffer, short width, short height, short new_depth, short dither )
{
	short i,j,r,g,b;
	unsigned char *iptr,*optr;
	iptr = optr = (unsigned char *)buffer;

	color_IndexInit();

	//	
	// Dither Image to 1-bit
	//

	if (new_depth==1) {
		unsigned long val = 0;

		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {
				r = (unsigned char)iptr[0];
				g = (unsigned char)iptr[1];
				b = (unsigned char)iptr[2];
				iptr += 3;

				color_ConvertPixel( &r,&g,&b, new_depth, dither,i,j );

				if (g<127) {
					*optr |=   128 >> (i&7);
				} else {
					*optr &= ~(128 >> (i&7));
				}

				if ((i&7)==7) *(++optr) = 0;
			}

			if (i&7) *(++optr) = 0;
			if (((width*new_depth+7)/8)&1) *(++optr) = 0;	// pad to short
		}
	}

	//	
	// Dither Image to 2-bits
	//

	else if (new_depth==2) {
		unsigned long val = 0;

		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {
				r = (unsigned char)iptr[0];
				g = (unsigned char)iptr[1];
				b = (unsigned char)iptr[2];
				iptr += 3;

				color_ConvertPixel( &r,&g,&b, new_depth, dither,i,j );

				*optr &= ~(0xc0 >> ((i&3)<<1) );
				*optr |=  ((4*255-r-g-g-b)>>8) << (6-((i&3)<<1));

				if ((i&3)==3) *(++optr) = 0;
			}

			if (i&3) *(++optr) = 0;
			if (((width*new_depth+7)/8)&1) *(++optr) = 0;	// pad to short
		}
	}

	//	
	// Dither Image to 4-bits
	//

	else if (new_depth==4) {
		unsigned long val = 0;

		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {
				r = (unsigned char)iptr[0];
				g = (unsigned char)iptr[1];
				b = (unsigned char)iptr[2];
				iptr += 3;

				color_ConvertPixel( &r,&g,&b, new_depth, dither,i,j );

				*optr &= (i&1)?0xf0:0x0f;
				*optr |=  ((4*255-r-g-g-b)>>6) << ((i&1)?0:4);

				if ((i&1)==1) *(++optr) = 0;
			}

			if (i&1) *(++optr) = 0;
			if (((width*new_depth+7)/8)&1) *(++optr) = 0;	// pad to short
		}
	}

	//
	// Remap image to 8 bits
	//

	else if (new_depth==8) {

		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {

				b = (unsigned char)*iptr++;
				g = (unsigned char)*iptr++;
				r = (unsigned char)*iptr++;

				color_ConvertPixel( &r,&g,&b, new_depth, dither,i,j );

				*optr++ = (unsigned char)color_IndexFromRGB( i,j,r,g,b );
			}

			if (width&1) *optr++ = 0;
		}
	}

	//
	// Remap image to 16 bits
	//

	else if (new_depth==16) {
		short tshort;

		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {

				b = (unsigned char)*iptr++;
				g = (unsigned char)*iptr++;
				r = (unsigned char)*iptr++;

				color_ConvertPixel( &r,&g,&b, new_depth, dither,i,j );

				tshort = ((r<<8)&0xf800) | ((g<<3)&0x07e0) | ((b>>3)&0x1f);
				*optr++ = tshort >> 8;
				*optr++ = tshort & 0xff;
			}
		}
	}

	return 0;
}
