/*======================================================================
:::
::: PaintLib.cpp
::: 
::: Platform-independent tealpaint image manipulation routines for 
::: the TealPoint image manager
:::
::: by TealPoint Software (c)1997-2003, All Rights Reserved
:::
======================================================================*/

//==========================================
// Includes
//==========================================

	#include <stdio.h>
#if !macintosh
	#include <windows.h>
#else
	#include <stdlib.h>
	#include <string.h>
#endif
	#include <math.h>

	#include "paintlib.h"
	#include "colorlib.h"
	#include "rlecode.h"

#if macintosh
	#define TRUE 1
	#define FALSE 0
#endif

#define MIN(a,b) ((a)<(b)?(a):(b))

//==========================================
// Prototypes
//==========================================

	static void MakePalette256( tRGBQuad *palette, int sr, int sg, int sb );

//==========================================
// Tables
//==========================================

	static int brite_remap4to8[] = {
		0,224,223,25,222,221,50,220,219,165,218,217,190,216,215,255
	};

	static int brite_remap2to8[] = {
		0,221,218,255
	};
	static int pl_gray_scale [] = { 34,68,85,119,136,170,187,221 };
	static int pl_color_scale[] = { 17,51,102,153,204,255 };

	static int pl_r_off[] = { 0, 18, 36, 54, 72, 90 };
	static int pl_g_off[] = { 0,  1,  2,  3,  4,  5 };
	static int pl_b_off[] = { 0,  6, 12,108,114,120 };


/*======================================================================================
:::
::: pnt_ImageGetInfo()
:::
::: Get information about an image in a TealPaint format database
:::
======================================================================================*/

short pnt_ImageGetInfo( FILE *i_fp, short *images, short index, char *title, short *rec, short *rows, short *cols, short *width, short *height, short *depth, short *layers, long *size )
{
	short record, records;
	char name[1000];
	short row,col,twidth,theight,tlayer,tlayers = 0;
	short rec_width, rec_height, rec_depth;
	short cache_cols = 1, cache_rows = 1, cache_record = 0, cache_images = 0, cache_depth = 1, cache_layers = 1, cache_width=160, cache_height=160;
	long rec_size;
	short error = -1;

	fseek( i_fp, COUNT_OFFSET, SEEK_SET );
	records = ReadMShort( i_fp );

	//==============================================
	// Scan thru records
	//==============================================
	
	for (record=0; record<records && (images || index>=-1); record++) {

		pnt_RecordInfo( i_fp, record, name, &rec_width, &rec_height, &rec_depth, &rec_size, NULL, NULL );
		
		twidth = theight = tlayer = tlayers = 0;

		if ( sscanf( name, "#(%hd,%hd) - %hdx%hd {%hd/%hd} ", &col, &row, &twidth, &theight, &tlayer, &tlayers ) < 4 ) {
			row=col=0;
		}
							
		// If this record is a start of image
							
		if ( !row && !col && tlayer<=1 ) {
			cache_images++;
			if (!index--) {

				if (size) *size = rec_size;
						
				cache_record  = record;
				cache_depth   = rec_depth;

				//--------------------------------------
				// Old single-record format
				//--------------------------------------
				
				if (!twidth || !theight) {
					cache_width  = rec_width;
					cache_height = rec_height;
					cache_cols   = 1;
					cache_rows   = 1;
					
					if (title) strcpy( title, name );
					index--;
				}

				//--------------------------------------
				// New multi-record format
				//--------------------------------------

				else {
					cache_width  = twidth;
					cache_height = theight;
					cache_rows = (theight+159)/160;
					cache_cols = (twidth +159)/160;

					if (tlayers) cache_layers = tlayers;
					if (title) {
						char *tstr;
						tstr = strstr( strstr( name, "x"), " " )+1;
						if (tstr[0]=='{' && tlayers) tstr = strstr( tstr, " " )+1;
						strcpy( title, tstr );
					}
				}

				error = 0;
			}
		}
		else if (index==-1) {
			if (size) *size += rec_size;
   	}
	}

	if (!cache_cols) cache_cols = (cache_width +159)/160;
	if (!cache_rows) cache_rows = (cache_height+159)/160;

	if (rec   ) *rec    = cache_record;
	if (images) *images = cache_images;
	if (rows  ) *rows   = cache_rows;
	if (cols  ) *cols   = cache_cols;
	if (width ) *width  = cache_width;
	if (height) *height = cache_height;
	if (depth ) *depth  = cache_depth;
	if (layers) *layers = cache_layers;

	if (index<0) return 0;

	return error;
}


/*======================================================================
:::
::: pnt_ImageRead()
:::
::: Read a full image and return and RGB buffer with the data
:::
======================================================================*/

short pnt_ImageRead( FILE *i_fp, short index, char **imagebuf, short *pwidth, short *pheight )
{
	short rec,rows,cols,width,height,layers,depth,m,n,q,twidth,theight,i,j;
	char *tbuf, *tptr, *sptr;

	pnt_ImageGetInfo( i_fp, NULL, index, NULL, &rec, &rows, &cols, &width, &height, &depth, &layers, NULL );
	if (pwidth ) *pwidth  = width;
	if (pheight) *pheight = height;
	if (rows>1 || cols>1 || layers>1) {

		if (*imagebuf = (char *)malloc( (long)width * height * 3 )) {

			for (n=0; n<rows; n++) for (m=0; m<cols; m++,rec++) for (q=0; q<layers; q++) {

				if (!pnt_RecordRead( i_fp, (short)(rec+q*rows*cols), &tbuf, &twidth, &theight )) {

					for (j=0; j<theight && n*160+j<height; j++) {
						tptr = *imagebuf + (width*(j+n*160) + m*160)*3;
						sptr = tbuf + (j*twidth)*3;
						if (!q) {
							memcpy( tptr, sptr, MIN(width-m*160,twidth)*3 );
						} else {
							for (i=0; i<twidth && m*160+i<width; i++) {
								if ((unsigned char)sptr[0]<0xf8 || (unsigned char)sptr[1]<0xf8 || (unsigned char)sptr[2]<0xf8) {
									tptr[0] = sptr[0];
									tptr[1] = sptr[1];
									tptr[2] = sptr[2];
								}
								tptr += 3;
								sptr += 3;
							}
						}
					}

					free( tbuf );
				}
			}
			return 0;
		}
		
		#if macintosh
			printf("Cannot allocate memory buffer (%ld bytes needed).\nTry increasing the memory allocated to this application.\n", (long)width*height*3 );
		#endif
	} else {
		return pnt_RecordRead( i_fp, rec, imagebuf, pwidth, pheight );
	}
	return PAINT_ERROR_IMAGEREAD_allocating_memory;
}


/*======================================================================================
:::
::: pnt_ImageAdd()
:::
::: Add a new image to the end of a TealPaint database
:::
======================================================================================*/

short pnt_ImageAdd( char *dbase_name, char *image_name, char *oimagebuf, short owidth, short oheight, short width, short height, short depth, short dither, short brightness, short contrast )
{
	char *thumbbuf_m, *thumbbuf_c;
	short error=0,r,g,b;
	FILE *i_fp, *o_fp;
	int c,i,j;
	int chop_x, chop_y;
	long k,size;
	long oldsize,records,s_modulus,t_modulus;

#if !macintosh
	char drive[ _MAX_DRIVE ];
	char dir[ _MAX_DIR ];
	char name[ _MAX_FNAME ];
	char ext[ _MAX_EXT ];
	char tfile[_MAX_FNAME + _MAX_EXT + 1];
#else
	char tfile[255];
#endif
	char *body_ptr,*sptr,*tptr,*imagebuf;
	char body_compression = 0;
	long body_size,new_size;

	//======================================
	// Scale image to target size
	//======================================

	if (color_GetScaledBitmapRGB( &imagebuf, width, height, oimagebuf, owidth, oheight )) {
		return PAINT_ERROR_ADD_scaling_image;
	}

	//========================================
	// Adjust brightness/contrast
	//========================================

	tptr = imagebuf;
	for (k=width*height; k--;) {
		r = (unsigned char)tptr[0];
		g = (unsigned char)tptr[1];
		b = (unsigned char)tptr[2];
		color_AdjustBrightnessContrast( &r,&g,&b, brightness, contrast );
		*tptr++ = (char)r;
		*tptr++ = (char)g;
		*tptr++ = (char)b;
	}

	//========================================
	// Make Thumbnails
	//========================================

	if (!color_GetScaledBitmapRGB( &thumbbuf_m, THUMB_WIDTH, THUMB_HEIGHT, imagebuf, width, height )) {
		if (depth < 8 || !color_GetScaledBitmapRGB( &thumbbuf_c, THUMB_WIDTH, THUMB_HEIGHT, imagebuf, width, height )) {

			if (!color_ReduceRgbBitmap( thumbbuf_m, THUMB_WIDTH, THUMB_HEIGHT, 1, TRUE )) {
				if (depth < 8 || !color_ReduceRgbBitmap( thumbbuf_c, THUMB_WIDTH, THUMB_HEIGHT, 8, FALSE )) {

					//========================================
					// Convert body image
					//========================================

					if (!color_ReduceRgbBitmap( imagebuf, width, height, depth, dither )) {

						//========================================
						// Get temporary file name
						//========================================

#if macintosh
						strcpy( tfile, "tempfile.$$$" );
#else
						_splitpath( dbase_name, drive, dir, name, ext );
						strcpy( tfile, dbase_name );
						strcpy( tfile+strlen(tfile)-strlen(ext), "$$$" );
#endif

						//========================================
						// Allocate work buffers
						//========================================

						s_modulus = (width*depth+15)/16*2;
						t_modulus = (160*depth+15)/16*2;

						if (body_ptr = (char *)malloc( t_modulus * 160 * 2 + 100 )) {

							//========================================
							// Create output records (160x160)
							//========================================

							for (chop_y=0; chop_y < height && !error; chop_y += 160 ) {
								for (chop_x=0; chop_x < width && !error; chop_x += 160 ) {

									//----------------------------------
									// Copy data to create body block
									//----------------------------------

									memset( body_ptr, depth>8 ? 0xff:0, t_modulus * 160 );

									tptr = body_ptr;
									sptr = imagebuf + (long)chop_y * s_modulus + (long)chop_x*depth/8;

									for (j=chop_y; j<160+chop_y && j<height; j++) {

										memcpy( tptr, sptr, MIN(t_modulus,s_modulus-(long)chop_x*depth/8) );
										sptr += s_modulus;
										tptr += t_modulus;
									}

									//----------------------------------
									// Compress it
									//----------------------------------

									body_size = t_modulus * 160;
									new_size = rle_Compress( (unsigned char *)body_ptr, (unsigned char *)body_ptr + body_size, (short)body_size );

									if (new_size && new_size < body_size) {
										memmove( body_ptr, body_ptr + body_size, new_size );
										body_size = new_size;
										body_compression = 1;
									}

									//----------------------------------
									// copy backup file
									//----------------------------------

									remove( tfile );
									rename( dbase_name, tfile );

									//----------------------------------
									// open old database file
									//----------------------------------

									if (i_fp=fopen( tfile, "rb" )) {

										fseek( i_fp, 0, SEEK_END );
										size = ftell( i_fp );
										fseek( i_fp, 0, SEEK_SET );

										if (o_fp=fopen( dbase_name, "wb" )) {

											//-----------------------------------------
											// Copy Database Header
											//-----------------------------------------

											for (k=0; k<HEADER_SIZE; k++) fputc( fgetc( i_fp ), o_fp );

											records = ReadMShort( i_fp );		// increment record count
											WriteMShort( o_fp, records+1 );
										
											// Copy each offset over, adding 8 to the offset of each

											for (i=0; i<records; i++) {
												long offset = ReadMLong( i_fp );
												WriteMLong( o_fp, offset+8 );
												WriteMLong( o_fp, ReadMLong( i_fp ));
											}

											// Write New offset

											WriteMLong( o_fp, size+8 );
											WriteMLong( o_fp, 0 );

											//-----------------------------------------
											// Write the rest of the file
											//-----------------------------------------

											while ((c=fgetc(i_fp))!=EOF) fputc(c,o_fp);
											oldsize = ftell(o_fp);

											//-----------------------------------------
											// Write New Image Header
											//-----------------------------------------

											WriteMLong( o_fp, IH_size );									// thumb offset
											WriteMLong( o_fp, THUMBSIZE );									// thumb size
											WriteMLong( o_fp, IH_size+THUMBSIZE+(depth>=8 ? THUMBSIZEC:0) );// body offset
											WriteMLong( o_fp, body_size );									// body size
											WriteMLong( o_fp, IH_size+THUMBSIZE+(depth>=8 ? THUMBSIZEC:0) + body_size );	// name offset
							
											fputc( depth>=8 ? TH_FLAG_colorthumb:0, o_fp );
											fputc( body_compression,   o_fp );

											WriteMShort( o_fp,160   );
											WriteMShort( o_fp,160   );
											WriteMShort( o_fp,depth );

											//-----------------------------------------
											// Write Thumbnail
											//-----------------------------------------

											WriteMShort( o_fp, THUMB_WIDTH   );			// thumb header
											WriteMShort( o_fp, THUMB_HEIGHT  );
											WriteMShort( o_fp, THUMB_MODULUS );
											WriteMShort( o_fp, 0 );
											WriteMLong ( o_fp, 0 );
											WriteMLong ( o_fp, 0 );

											fwrite( thumbbuf_m, THUMBSIZE-16,  1, o_fp );				// mono
											if (depth>=8) fwrite( thumbbuf_c, THUMBSIZEC, 1, o_fp );	// color

											//-----------------------------------------
											// Write Body
											//-----------------------------------------

											fwrite( body_ptr, body_size,1, o_fp );

											//-----------------------------------------
											// Write Name
											//-----------------------------------------

											fprintf( o_fp, "#(%d,%d) - %dx%d %s%c", chop_x/160, chop_y/160, width, height, image_name,0 );

											fclose(o_fp);
										} else {
											error = PAINT_ERROR_ADD_opening_database;
										}

										fclose(i_fp);

										if (!error) remove( tfile );

									} else {
										error = PAINT_ERROR_ADD_opening_tempfile;
									}

								}
							}
							free( body_ptr );
						} else error = PAINT_ERROR_ADD_allocating_work_buffer;
					} else error = PAINT_ERROR_ADD_converting_image;
				} else error = PAINT_ERROR_ADD_converting_color_icon;

			} else error = PAINT_ERROR_ADD_converting_mono_icon;
			free( thumbbuf_c );
		} else error = PAINT_ERROR_ADD_scaling_color_icon;
		free( thumbbuf_m );
	} else error = PAINT_ERROR_ADD_scaling_mono_icon;

	free( imagebuf );

	#if macintosh
		if (error==PAINT_ERROR_ADD_allocating_work_buffer) {
			printf("Cannot allocate memory buffer.\nTry increasing the memory allocated to this application.\n", (long)width*height*3 );
		}
	#endif

	return error;
}


/*======================================================================
:::
::: pnt_RecordInfo()
:::
::: Get information about a single low-level record in a TP database
:::
======================================================================*/

short pnt_RecordInfo( FILE *i_fp, short record, char *name, short *width, short *height, short *depth, long *size, long *bodyoffset, short *compression )
{
	long offset;
	short j;

	//-------------------------------
	// Read image record
	//-------------------------------

	fseek( i_fp, TABLE_OFFSET + (record<<3), SEEK_SET );
	offset = ReadMLong( i_fp );

	//-------------------------------
	// Get Name
	//-------------------------------

	if (name) {
		fseek( i_fp, offset + IH_nameoffset, SEEK_SET );
		fseek( i_fp, offset + ReadMLong(i_fp), SEEK_SET );
		fgets( name, 999, i_fp );

		// Convert odd chars

		for (j=0; name[j]; j++) {
			if (name[j]==10 || name[j]==13 || name[j]==9) { name[j] = ' '; break; }
		}
	}

	//-------------------------------
	// Get image size
	//-------------------------------

	if (width) {
		fseek( i_fp, offset + IH_width, SEEK_SET );
		*width = ReadMShort( i_fp );
	}

	if (height) {
		fseek( i_fp, offset + IH_height, SEEK_SET );
		*height = ReadMShort( i_fp );
	}

	if (depth) {
		long headsize;

		fseek( i_fp, offset + IH_headsize, SEEK_SET );
		headsize = ReadMLong( i_fp );
		if (headsize >= IH_depth+2) {
			fseek( i_fp, offset + IH_depth, SEEK_SET );
			*depth = ReadMShort( i_fp );
		} else {
			*depth = 1;
		}
	}

	//-------------------------------
	// Misc
	//-------------------------------

	if (compression) {
		fseek( i_fp, offset + IH_compression, SEEK_SET );
		*compression = fgetc( i_fp );
	}

	if (bodyoffset) {
		fseek( i_fp, offset + IH_bodyoffset, SEEK_SET );
		*bodyoffset = offset + ReadMLong(i_fp);
	}

	//-------------------------------
	// Get file size
	//-------------------------------

	if (size) {
		fseek( i_fp, offset + IH_bodysize, SEEK_SET );
		*size = ReadMLong( i_fp );
	}

	return 0;
}


/*======================================================================
:::
::: pnt_RecordRead()
:::
::: Read an image record and return and RGB buffer with the data
:::
======================================================================*/

short pnt_RecordRead( FILE *i_fp, short record, char **imagebuf, short *pwidth, short *pheight )
{
	int y,x,z,c,i,superwidth,lx;
	long bodyoffset;
	short compression, width,height,depth, index;
	static tRGBQuad palette[256];	
	char *tbuf;

	//=====================================
	// Get Parameters
	//=====================================

	pnt_RecordInfo( i_fp, record, NULL, &width, &height, &depth, NULL, &bodyoffset, &compression );
	if (pwidth ) *pwidth  = width;
	if (pheight) *pheight = height;

	if (*imagebuf = tbuf = (char *)malloc( (long)width * height * 3 )) {

		fseek( i_fp, bodyoffset, SEEK_SET );

		//=====================================
		// Draw 1-bit record
		//=====================================

		if (depth==1) {
			superwidth = (width+15)/16*16;

			if (compression) {

				rle_DecompressInit( i_fp );

				for (y=0; y<height; y++) {
					lx = -1;
					for (x=0; x<superwidth; x+=8) {

						c = rle_DecompressGetByte();
					
						for (i=0; i<8 && x+i<width; i++) {
		
							if (c & (0x80>>i)) {
								*(tbuf)++ = 0;
								*(tbuf)++ = 0;
								*(tbuf)++ = 0;
							} else {
								*(tbuf)++ = (char)255;
								*(tbuf)++ = (char)255;
								*(tbuf)++ = (char)255;
							}
						}
					}
				}
		
				rle_DecompressTerm();
		
			} else {
				for (y=0; y<height; y++) {
					for (x=0; x<superwidth; x+=8) {
		
						c = fgetc(i_fp);
		
						for (i=0; i<8 && x+i<width; i++) {
							if (c & (0x80>>i)) {
								*(tbuf)++ = 0;
								*(tbuf)++ = 0;
								*(tbuf)++ = 0;
							} else {
								*(tbuf)++ = (char)255;
								*(tbuf)++ = (char)255;
								*(tbuf)++ = (char)255;
							}
						}
					}
				}
			}
		}

		//=====================================
		// Draw 2-bit record
		//=====================================

		else if (depth==2) {

			MakePalette256( palette,255,255,255 );

			superwidth = (width+3)&~3L;

			if (compression) {

				rle_DecompressInit( i_fp );

				for (y=0; y<height; y++) {
					lx = -1;
					for (x=0; x<superwidth; ) {

						c = (unsigned char)rle_DecompressGetByte();

						for (z=0; z<4; z++,c<<=2,x++) {

							if (x<width) {
								index = brite_remap2to8[(c>>6)&3];
								*(tbuf)++ = palette[index].red;
								*(tbuf)++ = palette[index].green;
								*(tbuf)++ = palette[index].blue;
							}
						}
					}
					if (superwidth&1) rle_DecompressGetByte();
				}	
		
				rle_DecompressTerm();
			} 
			
			else {
				for (y=0; y<height; y++) {
					for (x=0; x<superwidth;) {
		
						c = fgetc(i_fp);
		
						for (z=0; z<4; z++,c<<=2,x++) {

							if (x<width) {
								index = brite_remap2to8[(c>>6)&3];
								*(tbuf)++ = palette[index].red;
								*(tbuf)++ = palette[index].green;
								*(tbuf)++ = palette[index].blue;
							}
						}
					}
				}
			}
		}

		//=====================================
		// Draw 4-bit record
		//=====================================

		else if (depth==4) {

			MakePalette256( palette,255,255,255 );

			superwidth = (width+1)&~1L;

			if (compression) {

				rle_DecompressInit( i_fp );

				for (y=0; y<height; y++) {
					lx = -1;
					for (x=0; x<superwidth; ) {

						c = (unsigned char)rle_DecompressGetByte();

						for (z=0; z<2; z++,c<<=4,x++) {

							if (x<width) {
								index = brite_remap4to8[(c>>4)&15];
								*(tbuf)++ = palette[index].red;
								*(tbuf)++ = palette[index].green;
								*(tbuf)++ = palette[index].blue;
							}
						}
					}
					if (superwidth&1) rle_DecompressGetByte();
				}	
		
				rle_DecompressTerm();
			} else {
				for (y=0; y<height; y++) {
					for (x=0; x<superwidth;) {
		
						c = fgetc(i_fp);
		
						for (z=0; z<2; z++,c<<=4,x++) {

							if (x<width) {
								index = brite_remap4to8[(c>>4)&15];
								*(tbuf)++ = palette[index].red;
								*(tbuf)++ = palette[index].green;
								*(tbuf)++ = palette[index].blue;
							}
						}
					}
				}
			}
		}


		//=====================================
		// Draw 16-bit record
		//=====================================

		else if (depth==16) {
			MakePalette256( palette,255,255,255 );

			superwidth = (width+1)&~1L;

			if (compression) {

				rle_DecompressInit( i_fp );

				for (y=0; y<height; y++) {
					lx = -1;
					for (x=0; x<superwidth; x++) {

						c  = rle_DecompressGetByte();
						c <<= 8;
						c |= rle_DecompressGetByte();
					
						if (x<width) {
							*(tbuf)++ = (c>>8) & 0xf8;
							*(tbuf)++ = (c>>3) & 0xfc;
							*(tbuf)++ = (c<<3) & 0xf8;
						}
					}
				}
		
				rle_DecompressTerm();
		
			} else {
				for (y=0; y<height; y++) {
					for (x=0; x<superwidth; x++) {
		
						c   = fgetc(i_fp);
						c <<= 8;
						c  |= fgetc(i_fp);
		
						if (x<width) {
							*(tbuf)++ = (c>>8) & 0xf8;
							*(tbuf)++ = (c>>3) & 0xfc;
							*(tbuf)++ = (c<<3) & 0xf8;
						}
					}
				}
			}
		}	

		//=====================================
		// Draw 8-bit record
		//=====================================

		else {
			MakePalette256( palette,255,255,255 );

			superwidth = (width+1)&~1L;

			if (compression) {

				rle_DecompressInit( i_fp );

				for (y=0; y<height; y++) {
					lx = -1;
					for (x=0; x<superwidth; x++) {

						c = rle_DecompressGetByte();
					
						if (x<width) {
							*(tbuf)++ = palette[c].red;
							*(tbuf)++ = palette[c].green;
							*(tbuf)++ = palette[c].blue;
						}
					}
				}
		
				rle_DecompressTerm();
		
			} else {
				for (y=0; y<height; y++) {
					for (x=0; x<superwidth; x++) {
		
						c = fgetc(i_fp);
		
						if (x<width) {
							*(tbuf)++ = palette[c].red;
							*(tbuf)++ = palette[c].green;
							*(tbuf)++ = palette[c].blue;
						}
					}
				}
			}
		}	
		return 0;
	}

	#if macintosh
		printf("Cannot allocate memory buffer.\nTry increasing the memory allocated to this application.\n", (long)width*height*3 );
	#endif
	return -1;
}


/***************************************************************************************
:::
::: pnt_DatabaseNew()
:::
::: Create a new blank TealPaint database file
:::
***************************************************************************************/

	static unsigned char data_header[HEADER_SIZE] = {
		'N', 'e', 'w', 'D', 'a', 't', 'a', 'b',
		'a', 's', 'e', 0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,   // ........
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,   // ........
		0x00,0x08,0x00,0x00,0x33,0xb8,0xbd,0xa5,   // ....3..#	attr.w ver.w creat_date.l
		0x33,0xb8,0xbd,0xa5,0x33,0xb8,0xbd,0xa5,   // 3...3...  mod_date.l back_date.l
		0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,   // ........  mod_num.l app_info.l
		0x00,0x00,0x00,0x00,0x44,0x61,0x74,0x61,   // ....Data	app_sort.l type.l
		0x54,0x6c,0x50,0x74,0x00,0x00,0x00,0x00,   // TlPt....	creator.l id_seed.l

		0x00,0x00,0x00,0x00,					   // ....			record list type -> first header
	};

short pnt_DatabaseNew( char *out_file )
{
	FILE *o_fp;
	int count = 0,i;
#if !macintosh
	char drive	[ _MAX_DRIVE ];
	char dir	[ _MAX_DIR ];
	char fname	[ _MAX_FNAME ];
	char ext	[ _MAX_EXT ];
#endif

#if !macintosh
	_splitpath( out_file, drive, dir, fname, ext );
	strncpy( (char *)data_header, fname, 29 );
#else
	strncpy( (char *)data_header, out_file, 29 );
#endif

	if (o_fp = fopen( out_file, "wb")) {

		//
		// Output database header
		//

		fwrite ( data_header, HEADER_SIZE, 1, o_fp );	// record
		WriteMShort( o_fp, count );

		//
		// Output temporary record headers (8-bytes each)
		//

		for (i=0; i<count; i++) {
			WriteMLong( o_fp, 0 );
			WriteMLong( o_fp, 0 );
		}

		//
		// Output middle weird stuff (is this high byte of record count?  Don't know!)
		//

		fputc  ( 0, o_fp );
		fputc  ( 0, o_fp );
			
		fclose( o_fp );

		return 0;
	}

	return PAINT_ERROR_DBASENEW_creating_database;
}


/***************************************************************************************
:::
::: pnt_BmpRead()
:::
::: read a bmp file and allocate it into a 24-bit pixel array
:::
***************************************************************************************/

	static int bmp_loadmask[] = { 0,7,3,0,1, 0,0,0,0   };
	static int bmp_shift   [] = { 0,7,6,0,4, 0,0,0,0   };
	static int bmp_readmask[] = { 0,1,3,0,15,0,0,0,255 };

short pnt_BmpRead( char *file, long *wptr, long *hptr, long *dptr, char **bufptr, int mode )
{
	long offset;
	char *buffer, *tptr;
	int c,i,j,count, width, height, depth, owidth, oheight;
	int in_mod, out_mod;
	FILE *i_fp;
	BMP_HEADER header;
	tRGBQuad palette[256];
	char tstr[2];

	if (!(i_fp = fopen( file, "rb"))) {
		return PAINT_ERROR_BMPREAD_opening_image;
	}

	tstr[0] = fgetc( i_fp );
	tstr[1] = fgetc( i_fp );
 		
	if ( tstr[0] != 'B' || tstr[1] != 'M' ) {
		fclose( i_fp );
		return PAINT_ERROR_BMPREAD_incorrect_format;
	}

	// Save offset to bitmap data

	ReadILong(i_fp);
	ReadILong(i_fp);
	offset = ReadILong(i_fp);

	//
	// Read Header
	//

	header.biSize			=	ReadILong( i_fp );
	header.biWidth 			=	ReadILong( i_fp );
	header.biHeight			=	ReadILong( i_fp );
	header.biPlanes			=	ReadIShort( i_fp );
	header.biBitCount		=	ReadIShort( i_fp );
	header.biCompression	=	ReadILong( i_fp );
	header.biSizeImage		=	ReadILong( i_fp );
	header.biXPelsPerMeter	=	ReadILong( i_fp );
	header.biYPelsPerMeter	=	ReadILong( i_fp );
	header.biClrUsed		=	ReadILong( i_fp );
	header.biClrImportant	=	ReadILong( i_fp );

	width  = owidth  = header.biWidth;
	height = oheight = header.biHeight;
	depth  = header.biBitCount;

	if (owidth  < 160) owidth = 160;

	owidth = (owidth+15)&~15L;	// round up to multiple of 16

	if (oheight < 144) oheight = 144;

	if (depth!=1 && depth!=2 && depth!=4 && depth!=8 && depth!=24) {
		fclose(i_fp);
		return PAINT_ERROR_BMPREAD_incorrect_format;
	}

	in_mod  = ((width * depth + 31)/32)*4;
	out_mod =  owidth * 3;

	//======================================
	// Allocate 24-bit image buffer
	//======================================

	buffer = (char *)malloc( oheight * out_mod );

	if (buffer) {
		memset( buffer, 255, oheight * out_mod );

		// Read Palette

		if (header.biClrUsed) {
			fread( palette, header.biClrUsed<256 ? header.biClrUsed : 256, 4, i_fp );
		} else if (depth<=8) {
			fread( palette, 1<<depth, 4, i_fp );
		}

		// Seek to bitmap data

		fseek( i_fp, offset, SEEK_SET );

		switch( depth ) {

			//======================================
			// Palette code
			//======================================

			case 1:
			case 2:
			case 4:
			case 8:

				for (j=0; j<height; j++) {
					tptr = &buffer[out_mod*(height-j-1)];
					count = in_mod;
					for (i=0; i<width; i++) {

						if (!(i&bmp_loadmask[depth])) {
							c = fgetc( i_fp );
							count--;
						} else {
							c <<= depth;
						}

						tptr[0] = palette[(c>>bmp_shift[depth]) & bmp_readmask[depth]].blue;
						tptr[1] = palette[(c>>bmp_shift[depth]) & bmp_readmask[depth]].green;
						tptr[2] = palette[(c>>bmp_shift[depth]) & bmp_readmask[depth]].red;
						tptr += 3;
					}

					for (i=width; i<owidth; i++) {
						*tptr++ = (char)255;
						*tptr++ = (char)255;
						*tptr++ = (char)255;
					}

					if (count>0) fseek( i_fp, count, SEEK_CUR );
				}
				break;

			//======================================
			// 24-bit code
			//======================================

			case 24:
				for (j=0; j<height; j++) {
					tptr = &buffer[out_mod*(height-j-1)];
					count = in_mod;
					for (i=0; i<width; i++) {

						tptr[0] = fgetc( i_fp );
						tptr[1] = fgetc( i_fp );
						tptr[2] = fgetc( i_fp );
						tptr += 3;

						count -= 3;
					}

					for (i=width; i<owidth; i++) {
						*tptr++ = (char)0xff;
						*tptr++ = (char)0xff;
						*tptr++ = (char)0xff;
					}

					if (count>0) fseek( i_fp, count, SEEK_CUR );
				}
				break;

		}
	
	} else {
		#if macintosh
			printf("Cannot allocate memory buffer.\nTry increasing the memory allocated to this application.\n", (long)width*height*3 );
		#endif
		return PAINT_ERROR_BMPREAD_allocating_memory;
	}

	fclose( i_fp );

	*bufptr = buffer;
	*wptr = owidth;
	*hptr = oheight;
	*dptr = depth;
	return 0;
}


/***************************************************************************************
:::
::: pnt_BmpWrite()
:::
::: write a bmp file from a 24-bit pixel array
:::
***************************************************************************************/

short pnt_BmpWrite( char *oname, char *imagebuf, short width, short height, short depth )
{
	BMP_HEADER header;
	long bodysize = (long)(width*3+3)/4*4 * height;
	short palette_size = 0,i,j;
	FILE *o_fp;
	char *tbuf;

	if (!(o_fp = fopen( oname, "wb" ))) {
		return PAINT_ERROR_BMPWRITE_opening_image;
	}

	//=======================================
	// Fill Header
	//=======================================

	header.biSize   = sizeof(BMP_HEADER);
	header.biWidth  = width;
	header.biHeight = height;
	header.biPlanes = 1;
	header.biBitCount      = 24;
	header.biCompression   = 0;
	header.biSizeImage     = 0;
	header.biXPelsPerMeter = 2667;
	header.biYPelsPerMeter = 2667;
	header.biClrUsed       = 0;
	header.biClrImportant  = 0;

	//=======================================
	// Write Header
	//=======================================

	fputc( 'B', o_fp );
	fputc( 'M', o_fp );

	WriteILong ( o_fp, bodysize + sizeof(BMP_HEADER) + 14 + 4*palette_size );
	WriteILong ( o_fp, 0 );
	WriteILong ( o_fp, sizeof(BMP_HEADER) + 14 + 4*palette_size );  // offset to body

	WriteILong ( o_fp, header.biSize );
	WriteILong ( o_fp, header.biWidth );
	WriteILong ( o_fp, header.biHeight );
	WriteIShort( o_fp, header.biPlanes );
	WriteIShort( o_fp, header.biBitCount );
	WriteILong ( o_fp, header.biCompression );
	WriteILong ( o_fp, header.biSizeImage );
	WriteILong ( o_fp, header.biXPelsPerMeter );
	WriteILong ( o_fp, header.biYPelsPerMeter );
	WriteILong ( o_fp, header.biClrUsed );
	WriteILong ( o_fp, header.biClrImportant );

	//=======================================
	// write out palette
	//=======================================

	for (i=0; i<palette_size; i++) WriteILong( o_fp, i ? 0:0xffffffff );

	//=======================================
	// write out body
	//=======================================

	for (j=0; j<height; j++) {
		tbuf = imagebuf + width*(height-j-1)*3;
		for (i=0; i<width; i++) {

			fputc( tbuf[2], o_fp );
			fputc( tbuf[1], o_fp );
			fputc( tbuf[0], o_fp );
			tbuf+=3;
		}
		for (i*=3;i&3; i++) fputc( 0xff, o_fp );
	}


	fclose( o_fp );

	return 0;
}


/*---------------------------------------------------------------------------
:::
::: MakePalette256()
:::
---------------------------------------------------------------------------*/

static void MakePalette256( tRGBQuad *palette, int sr, int sg, int sb )
{
	int i,r,g,b;
	for (r=0;r<6;r++) for (g=0;g<6;g++) for (b=0;b<6;b++) {
		palette[ pl_r_off[r] + pl_g_off[g] + pl_b_off[b] ].red   = (sr*pl_color_scale[5-r]/255);
		palette[ pl_r_off[r] + pl_g_off[g] + pl_b_off[b] ].green = (sg*pl_color_scale[5-g]/255);
		palette[ pl_r_off[r] + pl_g_off[g] + pl_b_off[b] ].blue  = (sb*pl_color_scale[5-b]/255);
	}

	for (i=6*6*6; i<256; i++) {
		if (i<6*6*6+8) {
			palette[i].red   = (sr*pl_gray_scale[(i-6*6*6)]/255); 
			palette[i].green = (sg*pl_gray_scale[(i-6*6*6)]/255);
			palette[i].blue  = (sb*pl_gray_scale[(i-6*6*6)]/255);
		} else if (i==6*6*6+8) {
			palette[i].red = (sr*238/255); palette[i].green = (sg*238/255); palette[i].blue = (sb*238/255);
		} else if (i==6*6*6+9) {                                                             
			palette[i].red = (sr*204/255); palette[i].green = (sg*204/255); palette[i].blue = (sb*204/255);
		} else if (i==6*6*6+10) {                                                            
			palette[i].red = (sr*136/255); palette[i].green = (sg*  0/255); palette[i].blue = (sb*  0/255);
		} else if (i==6*6*6+11) {                                                            
			palette[i].red = (sr*136/255); palette[i].green = (sg*  0/255); palette[i].blue = (sb*136/255);
		} else if (i==6*6*6+12) {                                                            
			palette[i].red = (sr*  0/255); palette[i].green = (sg*136/255); palette[i].blue = (sb*  0/255);
		} else if (i==6*6*6+13) {                                                            
			palette[i].red = (sr*  0/255); palette[i].green = (sg*136/255); palette[i].blue = (sb*136/255);
		} else {
			palette[i].red = 0; palette[i].green = 0; palette[i].blue = 0;
		}
	}
}

/***************************************************************************************
:::
::: Endian File Routines
:::
***************************************************************************************/

void WriteILong(FILE *fp,unsigned long value)
{
	fputc( value,    fp );
	fputc( value>>8, fp );
	fputc( value>>16,fp );
	fputc( value>>24,fp );
}

void WriteMLong(FILE *fp,unsigned long value)
{
	fputc( value>>24,fp );
	fputc( value>>16,fp );
	fputc( value>>8, fp );
	fputc( value,    fp );
}

void WriteMShort(FILE *fp,unsigned long value)
{
	fputc( value>>8, fp );
	fputc( value,    fp );
}

void WriteIShort(FILE *fp,unsigned long value)
{
	fputc( value,    fp );
	fputc( value>>8, fp );
}

long ReadMLong(FILE *fp)
{
	long value;
	value = (unsigned char)fgetc( fp );
	value = (value<<8) | (unsigned char)fgetc( fp );
	value = (value<<8) | (unsigned char)fgetc( fp );
	value = (value<<8) | (unsigned char)fgetc( fp );
	return value;
}

long ReadILong(FILE *fp)
{
	long value;
	value  =  (unsigned char)fgetc( fp );
	value |= (long)(unsigned char)fgetc( fp ) <<8;
	value |= (long)(unsigned char)fgetc( fp ) <<16;
	value |= (long)(unsigned char)fgetc( fp ) <<24;
	return value;
}

short ReadMShort(FILE *fp)
{
	short value;
	value = fgetc( fp );
	value = (value<<8) | fgetc( fp );
	return value;
}
short ReadIShort(FILE *fp)
{
	short value;
	value = fgetc( fp );
	value |= (fgetc( fp )<<8);
	return value;
}


