/*======================================================================
:::
::: Paintlib.h
::: 
::: Platform-independent tealpaint image manipulation routines for 
::: the TealPoint image manager
:::
::: by TealPoint Software (c)1997-2003, All Rights Reserved
:::
======================================================================*/

//==============================================
// Defines
//============================================== 

	#define HEADER_SIZE (9*8+4)
	#define	COUNT_OFFSET	0x4C
	#define	TABLE_OFFSET	0x4E
	#define	DBASEID_OFFSET	0x40

	// TealPaint Image format

	#define IH_size			28

	#define	IH_thumboffset	 	0
	#define	IH_thumbsize		4
	#define	IH_bodyoffset		8
	#define	IH_bodysize			12
	#define	IH_nameoffset		16

	#define	IH_flags			20
	#define	IH_compression		21

	#define	IH_width			22
	#define	IH_height			24

	#define	IH_depth	    	26
	#define	IH_headsize	    	IH_thumboffset

	#define TH_FLAG_template	1	// means name is followed by a :<template database name>:<image number>
	#define TH_FLAG_memo		2	// means memo field is stored between image and name
	#define TH_FLAG_colorthumb	4	// means thumbnail is followed by an 8-bit thumbnail

	#define THUMB_HEIGHT	18
	#define THUMB_WIDTH		20
	#define THUMB_MODULUS	4
	#define THUMBSIZE	(THUMB_MODULUS * THUMB_HEIGHT + 16)
	#define THUMBSIZEC	(THUMB_WIDTH * THUMB_HEIGHT)


	// Misc

	#define PAINT_FLAG_transparent	1

	// Errors

	#define PAINT_ERROR_ADD_opening_database			-0x10
	#define PAINT_ERROR_ADD_opening_tempfile			-0x11
	#define PAINT_ERROR_ADD_allocating_work_buffer		-0x12
	#define PAINT_ERROR_ADD_converting_image			-0x13
	#define PAINT_ERROR_ADD_converting_color_icon		-0x14
	#define PAINT_ERROR_ADD_converting_mono_icon		-0x15
	#define PAINT_ERROR_ADD_scaling_color_icon			-0x16
	#define PAINT_ERROR_ADD_scaling_mono_icon			-0x17
	#define PAINT_ERROR_ADD_scaling_image				-0x18
 
	#define PAINT_ERROR_BMPWRITE_opening_image			-0x20

	#define PAINT_ERROR_BMPREAD_opening_image			-0x30
	#define PAINT_ERROR_BMPREAD_incorrect_format		-0x31
	#define PAINT_ERROR_BMPREAD_allocating_memory		-0x32

	#define PAINT_ERROR_IMAGEREAD_allocating_memory		-0x40

	#define PAINT_ERROR_DBASENEW_creating_database		-0x50


//==============================================
// Structs
//============================================== 

	typedef struct
	{
		long	biSize;
		long	biWidth;
		long	biHeight;
		short	biPlanes;
		short 	biBitCount;
		long	biCompression;
		long	biSizeImage;
		long	biXPelsPerMeter;
		long	biYPelsPerMeter;
		long	biClrUsed;
		long	biClrImportant;

	} BMP_HEADER;

	typedef struct {
		unsigned char blue;
		unsigned char green;
		unsigned char red;
		unsigned char reserved;
	} tRGBQuad;


//==============================================
// Prototypes
//============================================== 

	// Image routines

	short	pnt_ImageGetInfo( FILE *i_fp, short *images, short index, char *title, short *rec, short *rows, short *cols, short *width, short *height, short *depth, short *layers, long *size );
	short	pnt_ImageRead	( FILE *i_fp, short index, char **imagebuf, short *pwidth, short *pheight );
	short	pnt_ImageAdd	( char *dbase_name, char *image_name, char *imagebuf, short owidth, short oheight, short width, short height, short depth, short dither, short brite, short contr );

	// Record routines

	short	pnt_RecordInfo	( FILE *i_fp, short record, char *name, short *width, short *height, short *depth, long *size, long *bodyoffset, short *compression );
	short	pnt_RecordRead	( FILE *i_fp, short record, char **imagebuf, short *pwidth, short *pheight );

	// Database routines

	short	pnt_DatabaseNew	( char *out_file );

	// BMP routines

	short	pnt_BmpRead		( char *file, long *wptr, long *hptr, long *dptr, char **bufptr, int mode );
	short	pnt_BmpWrite	( char *oname, char *imagebuf, short width, short height, short depth );

	// Endian File Routines

	void	WriteILong		(FILE *fp,unsigned long value);
	void	WriteMLong		(FILE *fp,unsigned long value);
	void	WriteMShort		(FILE *fp,unsigned long value);
	void	WriteIShort		(FILE *fp,unsigned long value);
	long	ReadMLong		(FILE *fp);
	long	ReadILong		(FILE *fp);
	short	ReadMShort		(FILE *fp);
	short	ReadIShort		(FILE *fp);

	