//======================================================================
//
// PICUTIL.C
//
// Main code for Picutil TealPaint Image Utility
//
// (c)2000-2003 TealPoint Software, All Rights Reserved
//
//======================================================================

//=========================================
// Includes
//=========================================

	#include <stdio.h>
	#include <string.h>

//	#include "rlecode.h"
//	#include "colorlib.h"
//	#include "paintlib.h"

//=========================================
// Modules
//=========================================

	#include "rlecode.cpp"
	#include "colorlib.cpp"
	#include "paintlib.cpp"

//=========================================
// Prototypes
//=========================================

	void GetFile( char *name, char *extension, char *mode );

//=========================================
// Mac suckiness
//=========================================

	#if macintosh
	#define	STR_UpperCase(_c)		(((_c)>='a' && (_c)<='z') ? (_c)+'A'-'a' : (_c))

	int strcmpi( char *str1, char *str2 )
	{
		short i;
		for (i = 0; str1[i] && str2[i]; i++) {
			if (STR_UpperCase(str1[i]) != STR_UpperCase(str2[i])) return STR_UpperCase(str1[i])-STR_UpperCase(str2[i]);		// revisit and check sign
		}
		return (STR_UpperCase(str1[i])-STR_UpperCase(str2[i]));
	}
	
	#endif


/*======================================================================
:::
::: main()
:::
======================================================================*/

int main(int argc, char **argv)
{
	char menu = 0;
	char dbase_name[200];
	char bmp_name[200];
	char tstring[1000];
	int q_mac = 1, q_batch = 0;
	FILE *fp;
						
	printf("-----------------------------------------------------------\n"
			 " PICUTIL Image management utility for TealPoint Software \n"
			 " Version 6.00.01, %s\n"
			 " (c)1998-2004 TealPoint Software, All Rights Reserved \n"
			 "-----------------------------------------------------------\n",__DATE__);

	//==========================================
	// Detect macintosh (no .exe extension)
	//==========================================
	
	if (argc>=1) {
		int len=strlen(argv[0]);
		if (len>4 && strcmpi(&argv[0][len-4],".exe")) q_mac = 0;
	}
	
	//==========================================
	// Batch Command
	//==========================================
		
	if (argc>=2) {
		if (argv[1][0]=='-') {
			menu = argv[1][1];
		} else {
			menu = '?';
		}
		q_batch = 1;
	}

	//==========================================
	// Interactive Menu
	//==========================================

	if (!menu) {
restart:
		do {
			if (menu != ' ' && menu!=10 && menu !=13)
				printf("\n=================================================================\n   MENU:\n\n"
				 "   (L) = List images in a TealPaint Database\n"
				 "   (N) = Create a new TealPaint Database\n"
				 "   (E) = Export BMP image from a TealPaint Database\n"
				 "   (I) = Import BMP image to a TealPaint Database\n"
				 "%s"
 				 "  Please Enter Selection-> ",				 
				 q_mac ? "\n   (Choose 'Quit' from the 'File' Menu when done)\n\n" : "   (Q) = Quit\n\n" );
			
			scanf( "%c", &menu );
			
		} while (menu!='l' && menu!='L' &&
		         menu!='n' && menu!='N' &&
		         menu!='e' && menu!='E' &&
		         menu!='i' && menu!='i' &&
		         menu!='q' && menu!='Q'    );

		printf("\n");
		printf("=================================================================\n" );
	}

	//==========================================
	// Execute Commands
	//==========================================
	
	switch( menu ) {

		//==========================================
		// List Images
		//==========================================

		case 'l':
		case 'L':

			//-----------------------------------------
			// Get Database
			//-----------------------------------------
			
			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint-format .PDB database to list" );
			}

			//-----------------------------------------
			// List Database Contents
			//-----------------------------------------
			
			{
				short images = 0;
				short i, width, height, depth, layers;
				long size;

				if (fp = fopen( dbase_name, "rb" )) {				

					//-----------------------------------------
					// Get Image Count
					//-----------------------------------------
					
					if (!pnt_ImageGetInfo( fp, &images, -1, NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL )) {
						printf("\n   %hd Images(s) Found in Database\n\n", images);
					} else {
						printf("ERROR:  Database may be damaged or is not a TealPaint database file.\n\n");
					}
					
					//-----------------------------------------
					// List Images
					//-----------------------------------------

					for (i=0; i<images; i++) {
						if (!pnt_ImageGetInfo( fp, NULL, i, tstring, NULL, NULL, NULL, &width, &height, &depth, &layers, &size )) {
							printf("   %-4hd: %s\n         %hd-bit, %hd x %hd pixels, %hd layer(s), %ld bytes\n\n", i, tstring, depth, width, height, layers, size );
						}
					}					
					
					fclose( fp );
				} else {
					printf("ERROR:  Could not find or open database '%s'.\n\n", dbase_name );
				}
			}
			break;

		//==========================================
		// New Database
		//==========================================

		case 'n':
		case 'N':

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint-format .PDB database to create" );
			}

			if (!pnt_DatabaseNew( dbase_name )) {
				printf("Created Database %s.\n\n", dbase_name );
			} else {
				printf("ERROR: Cannot Create Database %s.  Check filename and available space.\n\n", dbase_name );
			}
			break;

		//==========================================
		// Export Image
		//==========================================

		case 'e':
		case 'E':

			//-----------------------------------------
			// Get Database and Bitmap Names
			//-----------------------------------------

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint-format .PDB database to read" );
			}
			if (argc >= 4)	{
				strcpy(bmp_name,argv[3]);
			} else {
				GetFile( bmp_name, ".BMP", ".BMP Bitmap file to create" );
			}

			//-----------------------------------------
			// Open Database
			//-----------------------------------------
			
			{
				short images = 0;
				short i, width, height, depth, layers;
				long size;

				if (fp = fopen( dbase_name, "rb+" )) {				

					//-----------------------------------------
					// Get Image Count
					//-----------------------------------------
					
					if (!pnt_ImageGetInfo( fp, &images, -1, NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL )) {
						printf("   %hd Images(s) Found in Database\n\n", images);
					} else {
						printf("ERROR:  Database may be damaged or is not a TealPaint database file.\n\n");
					}
					
					//-----------------------------------------
					// Get image index
					//-----------------------------------------

					if (images) {
						long image;
						char *imagebuf;
						short width,height;

						if (images>1) {
							printf( ">> Enter Image Number to Extract (0-%ld) -> ", (long)(images-1) );
							scanf( "%ld", &image ); 
						} else {
							image = 0;
						}

						//-----------------------------------------
						// Load pixels
						//-----------------------------------------

						if (!pnt_ImageRead( fp, (short)image, &imagebuf, &width, &height )) {

							//-----------------------------------------
							// Save BMP
							//-----------------------------------------

							if (!pnt_BmpWrite( bmp_name, imagebuf, width, height, 24 )) {
								printf("\nImage export successful (%hd x %hd pixels).\n\n", width, height );
							} else {
								printf("ERROR:  Could not write bmp file '%s'.\n\n", bmp_name );
							}
							
							free(imagebuf);
							
						} else {
							printf("ERROR:  Could export image %ld.\n\n", image );
						}

					}
										
					fclose( fp );
				} else {
					printf("ERROR:  Could not find or open database '%s'.\n\n", dbase_name );
				}
			}

			break;

		case 'i':
		case 'I':

			//-----------------------------------------
			// Get Database and Bitmap Names
			//-----------------------------------------

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint-format .PDB database to receive image" );
			}
			if (argc >= 4)	{
				strcpy(bmp_name,argv[3]);
			} else {
				GetFile( bmp_name, ".BMP", ".BMP Bitmap file to import" );
			}

			//-----------------------------------------
			// Open Database
			//-----------------------------------------
			
			{
				short images = 0;
				short i, width, height, depth, layers;
				long size;

				if (fp = fopen( dbase_name, "rb" )) {				

					//-----------------------------------------
					// Get Image Count
					//-----------------------------------------
					
					if (!pnt_ImageGetInfo( fp, &images, -1, NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL )) {
						printf("   %hd Images(s) Found in Database\n\n", images);
					} else {
						printf("ERROR:  Database may be damaged or is not a TealPaint database file.\n\n");
					}

					fclose( fp );					
					
					//-----------------------------------------
					// Get image index
					//-----------------------------------------

					if (images) {
						char *imagebuf;
						long owidth,oheight,odepth;
						short width, height, depth, dither, brite, contr;
						
						//-----------------------------------------
						// Load pixels
						//-----------------------------------------

						if (!pnt_BmpRead( bmp_name, &owidth, &oheight, &odepth, &imagebuf, 0 )) {

							printf("   BMP is %ld bits, %ld x %ld pixels\n\n", odepth, owidth, oheight);

							width  = (short)owidth;
							height = (short)oheight;
							depth  = 16;
							dither = 1;
							brite  = 0;
							contr  = 0;
							
							if (!q_batch) {

								gets( tstring ); 

								printf( ">> Enter target width (ENTER for %hd) -> ", width );
								gets( tstring ); 
								sscanf( tstring, "%hd", &width ); 

								height = (long)width * oheight/owidth;
								printf( ">> Enter target height (ENTER for %hd) -> ", height );
								gets( tstring ); 
								sscanf( tstring, "%hd", &height ); 

								printf( ">> Enter target bit depth (1,2,4,8, or 16, ENTER for %hd) -> ", depth );
								gets( tstring ); 
								sscanf( tstring, "%hd", &depth ); 

								printf( ">> Dither image (Y/n) -> " );
								gets( tstring );
								dither = !(tstring[0]=='n' || tstring[0]=='N'); 
      					}
							
							//-----------------------------------------
							// Save Image
							//-----------------------------------------

							if (!pnt_ImageAdd( dbase_name, bmp_name, imagebuf, (short)owidth, (short)oheight, width, height, depth, dither, brite, contr )) {
								printf("\nImage import successful (%hd x %hd x %hd pixels).\n\n", width, height, depth );
							} else {
								printf("ERROR:  Could not write to database file '%s'.\n\n", dbase_name );
							}

							free(imagebuf);

						} else {
							printf("ERROR:  Could not find or open bitmap '%s'.\n\n", bmp_name );
						}

					}
										

				} else {
					printf("ERROR:  Could not find or open database '%s'.\n\n", dbase_name );
				}
			}

			break;

		//==========================================
		// Show Help
		//==========================================

		default:
		
			printf("Usage: PICUTIL [<options>]\n"
				 "   where <options> =\n"
				 "      -l <.pdb file>                     = List images in TealPaint Database\n"
				 "      -n <.pdb file>                     = Create new TealPaint Database\n"
				 "      -e <.pdb file> <.bmp file> <index> = Export BMP image from TealPaint DB\n"
				 "      -i <.pdb file> <.bmp file>         = Import BMP image toTealPaint DB\n" );
			break;

		//==========================================
		// Quit
		//==========================================

		case 'q':
		case 'Q':
			return 0;
	}

	if (argc<2) goto restart;
	
	return 0;
}


/*======================================================================
:::
::: GetFile()
:::
======================================================================*/

void GetFile( char *name, char *extension, char *mode )
{
	printf( "----------------------------------------------------------------\n>> %s:\n>> Enter file name or full path if not same folder (%s)-> ", mode, extension );
	scanf( "%s", name ); 
	printf( "----------------------------------------------------------------\n" );
		
	// force extension
		
	if (strlen(name) < strlen(extension) ||
		strcmpi(name+strlen(name)-strlen(extension),extension )) {
		strcat(name,extension);
	}
}


#if 0
int mai_record = 0;

int strcmpi( char *s1, char *s2 )
{
	int c1,c2;
	do {
		c1=*s1++;
		c2=*s2++;
		if (c1>='a' && c1<='z') c1+= 'A'-'a';
		if (c2>='a' && c2<='z') c2+= 'A'-'a';
	} while (c1==c2 && c1 && c2);
	return (c1==c2) ? 0 : (c2>c1) ? 1 : -1;
}

void ShowError( char *error, char *title )
{
	printf("%s: %s\n",title,error );
}

int ListImage( FILE *fp, char version, char compression, int width, int height, int depth, 
		long  body_off, long body_size, char *name )
{
	(void)(fp,version,compression,body_off);
	
	printf( "%-2ld. (%ld x %ld) %ld bytes, %ld colors:  %s\n", (long)mai_record, (long)width, (long)height, body_size, (long)(1<<depth), name );
	return 0;
}


int main(int argc, char **argv)
{
	char menu = 0;
	char dbase_name[200];
	char bmp_name[200];
	int q_mac = 1;
					
	printf("-----------------------------------------------------\n"
			 " PICUTIL Image management utility for TealPaint  \n"
			 " Version 4.91.00, %s\n"
			 " (c)1998-2001 TealPoint Software, All Rights Reserved \n"
			 "-----------------------------------------------------\n",__DATE__);

	if (argc>=1) {
		int len=strlen(argv[0]);
		if (len>4 && !strcmpi(&argv[0][len-4],".exe")) q_mac = 0;
	}
	
	//
	// Interactive Code
	//
		
	if (argc>=2) {
		if (argv[1][0]=='-') {
			menu = argv[1][1];
		} else {
			menu = '?';
		}
	}

	if (!menu) {
restart:
		do {
			if (menu != ' ' && menu!=10 && menu !=13)
				printf("\nMenu:\n"
				 "   (L) = List images in .pdb file\n"
				 "   (N) = Create new .pdb file\n"
				 "   (X) = Extract image from .pdb file\n"
				 "   (E) = Extract image from .pdb file (b&w)\n"
				 "   (A) = Append image to .pdb file\n"
				 "   (C) = Append image to .pdb file (color or greyscale)\n"
//				 "   (G) = Append image to .pdb file (in 4-shade grey)\n"
//				 "   (Z) = Append image to .pdb file (in 16-shade grey)\n"
				 "%s"
 				 "  Please Enter Selection-> ",				 
				 q_mac ? "   (Choose 'Quit' from the 'File' Menu when done)\n" : "   (Q) = Quit\n\n" );
			
			scanf( "%c", &menu );
			
		} while (menu!='l' && menu!='L' &&
		         menu!='n' && menu!='N' &&
		         menu!='x' && menu!='X' &&
		         menu!='e' && menu!='E' &&
		         menu!='a' && menu!='A' &&
		         menu!='c' && menu!='C' &&
//		         menu!='g' && menu!='G' &&
//		         menu!='z' && menu!='Z' &&
		         menu!='q' && menu!='Q'    );

		printf("\n");
			
	}
	
	switch( menu ) {

		case 'l':
		case 'L':

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to list" );
			}
			
			{
				int records = 0;
				int i;

				if (!tpc_DatabaseInfo( dbase_name, &records, NULL, NULL, ShowError )) {

					printf("   %ld Record(s) Found\n\n",(long)records);

					for (i=0; i<records; i++) {
						mai_record = i;
						tpc_RecordInfo( dbase_name, i, ListImage, ShowError );
					}
				}
			}
			break;

		case 'n':
		case 'N':

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to create" );
			}

			if (!tpc_NewDatabase( dbase_name, ShowError )) {
				printf("Created Database %s.\n\n", dbase_name );
			}
			break;

		case 'e':
		case 'E':
		case 'x':
		case 'X':

			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to read" );
			}
			if (argc >= 4)	{
				strcpy(bmp_name,argv[3]);
			} else {
				GetFile( bmp_name, ".BMP", "Bitmap file to create" );
			}

			{
				int records = 0;

				if (!tpc_DatabaseInfo( dbase_name, &records, NULL, NULL, ShowError )) {

					if (records) {

						long image;
						if (argc >= 5) {
							sscanf( argv[4], "%ld", &image );
						} else {
							if (records > 1) {
								printf( "Enter Image Number to Extract (0-%ld) ->", (long)(records-1) );
								scanf( "%ld", &image ); 
							} else image=0;
						}
				
						if (!tpc_ExportImage( dbase_name, bmp_name, image, menu=='e' || menu=='E', ShowError )) {
							printf("Image %s created.\n\n", bmp_name );
						}

					} else {
						printf("ERROR: Database has no images\n");
					}
				}
			}
			break;

		case 'a':
		case 'A':
			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to append to" );
			}
			if (argc >= 4)	{
				strcpy(bmp_name,argv[3]);
			} else {
				GetFile( bmp_name, ".BMP", "Bitmap file to append" );
			}
			if (!tpc_ImportImage( dbase_name, bmp_name, 1, ShowError )) {
				printf("Added image %s to database %s.\n\n", bmp_name, dbase_name );
			}
			break;

		case 'c':
		case 'C':
			if (argc >= 3)	{
				strcpy(dbase_name,argv[2]);
			} else {
				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to append to" );
			}
			if (argc >= 4)	{
				strcpy(bmp_name,argv[3]);
			} else {
				GetFile( bmp_name, ".BMP", "Bitmap file to append" );
			}
			if (!tpc_ImportImage( dbase_name, bmp_name, 8, ShowError )) {
				printf("Added image %s to database %s.\n\n", bmp_name, dbase_name );
			}
			break;

//		case 'g':
//		case 'G':
//			if (argc >= 3)	{
//				strcpy(dbase_name,argv[2]);
//			} else {
//				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to append to" );
//			}
//			if (argc >= 4)	{
//				strcpy(bmp_name,argv[3]);
//			} else {
//				GetFile( bmp_name, ".BMP", "Bitmap file to append" );
//			}
//			if (!tpc_ImportImage( dbase_name, bmp_name, 2, ShowError )) {
//				printf("Added image %s to database %s.\n\n", bmp_name, dbase_name );
//			}
//			break;
//
//		case 'z':
//		case 'Z':
//			if (argc >= 3)	{
//				strcpy(dbase_name,argv[2]);
//			} else {
//				GetFile( dbase_name, ".PDB", "TealPaint .PDB file to append to" );
//			}
//			if (argc >= 4)	{
//				strcpy(bmp_name,argv[3]);
//			} else {
//				GetFile( bmp_name, ".BMP", "Bitmap file to append" );
//			}
//			if (!tpc_ImportImage( dbase_name, bmp_name, 4, ShowError )) {
//				printf("Added image %s to database %s.\n\n", bmp_name, dbase_name );
//			}
//			break;

		default:
		
			//
			// Help
			//
		
			printf("Usage: PICUTIL [<options>]\n"
				 "   where <options> =\n"
				 "      -l <.pdb file>                     = List images in .pdb file\n"
				 "      -n <.pdb file>                     = Create new .pdb file\n"
				 "      -x <.pdb file> <.bmp_name file> <index> = Extract image from .pdb*\n"
				 "      -e <.pdb file> <.bmp_name file> <index> = Extract image from .pdb (b&w)**\n"
				 "      -a <.pdb file> <.bmp_name file>         = Add image to .pdb (b&w) \n"
				 "      -c <.pdb file> <.bmp_name file>         = Add image to .pdb (color or greys)***\n"
//				 "      -g <.pdb file> <.bmp_name file>         = Add image to .pdb (4-shades)***\n"
//				 "      -z <.pdb file> <.bmp_name file>         = Add image to .pdb (16-shades)***\n"
				 "       * 1-bit images will be exported as black on blue-green\n"
				 "       ** 1-bit images will be exported as black and white\n"
				 "       *** viewable only on devices running OS 3.5\n" );
				 
			break;

		case 'q':
		case 'Q':
			return 0;
	}

	if (argc<2) goto restart;
	
	return 0;
}

#endif
