/*======================================================================
:::
::: RleCode.cpp
::: 
::: Run-length compression/decompression  routines for 
::: the TealPoint image manager
:::
::: by TealPoint Software (c)1997-2003, All Rights Reserved
:::
======================================================================*/

//==========================================
// Includes
//==========================================

	#include <stdio.h>
	#include "rlecode.h"

//==========================================
// Defines
//==========================================

	#define STATE_LIT		0
	#define STATE_RUN		1
	#define STATE_WAIT1	2
	#define STATE_WAIT2	3

//==========================================
// Globals
//==========================================

	static FILE          *msc_uncompress_fp;
	static int    		  msc_uncompress_state;
	static int  		  msc_uncompress_count;
	static unsigned char  msc_uncompress_val;

/*------------------------------------------------------
:::
::: rle_DecompressInit
:::
------------------------------------------------------*/
void rle_DecompressInit( FILE *fp )
{
	msc_uncompress_fp = fp;
	msc_uncompress_state = STATE_WAIT1;
}

/*------------------------------------------------------
:::
::: rle_DecompressTerm
:::
------------------------------------------------------*/
void rle_DecompressTerm( void )
{
}

/*------------------------------------------------------
:::
::: rle_DecompressGetByte
:::
------------------------------------------------------*/
unsigned char rle_DecompressGetByte( void )
{
	unsigned char retval,c;

	if ( msc_uncompress_state==STATE_WAIT1 ) {

		c = fgetc( msc_uncompress_fp );

		msc_uncompress_count = 1 + (c >> 1);

		if ( c & 1) {
			msc_uncompress_state = STATE_RUN;
			retval = msc_uncompress_val = fgetc( msc_uncompress_fp );
		} else {
			msc_uncompress_state = STATE_LIT;
			retval = fgetc( msc_uncompress_fp );
		}

	}

	else switch( msc_uncompress_state ) {

		case STATE_RUN:
			retval = msc_uncompress_val;
			break;

		case STATE_LIT:
			retval = fgetc( msc_uncompress_fp );
			break;
	}

	if (!--msc_uncompress_count) msc_uncompress_state = STATE_WAIT1;

	return retval;
}


/*========================================================
:::
::: rle_Compress
:::
========================================================*/

	#define MINRUN	3

long rle_Compress( unsigned char *source, unsigned char *data, short size )
{
	unsigned char *head;
	unsigned char *maxdata = data + size-1;
	unsigned char  last,val;
	long   count;
	short	 run_count;

	// Initialize

	count = 0;

	// Compress

	// State Wait2

wait2:
	if (!size--) goto done;
	last = *source++; 

	// State Wait1

wait1:
	if (!size--) goto done2;
	val = *source++; 

	if (val == last) {									// start run of two
		if (data+2 >= maxdata) return 0;
		head    = data++;
		head[0] = 3;
		*data++ = val;
		count += 2;
		goto run;

	} else {													// start literal of two

		if (data+3 >= maxdata) return 0;
		head    = data++;
		head[0] = 2;
		*data++ = last;
		*data++ = val;
		count += 3;
		run_count = 1;
		last = val;
		goto lit;

	}

	// State Run

run:
	if (!size--) goto done;
	val = *source++; 

	if (val == last) {						// increase run
		*head += 2;
		if (*head == 255) goto wait2;
 		else              goto run;
 	} else {													
		last = val;
		goto wait1;
	}

	// Literal

lit:
	if (!size--) goto done;
	val = *source++; 

	if (val == last) {
		run_count++;
	} else {
		run_count = 1;
	}
	last = val;

	if (run_count < MINRUN) {			// increase literal

		if (data+1 >= maxdata) return 0;
		*head  += 2;
		*data++ = val;			

		count++;

		if (*head == 254) goto wait2;
		else              goto lit;

	} else {										// backoff last two literals

		*head -= (MINRUN-1) << 1;
		data  -= MINRUN-1;

		if (data+2 >= maxdata) return 0;
		head    = data++;	// start run of three
		head[0] = 5;
		*data++ = val;
		goto run;
	}

done2:
		if (data+1 >= maxdata) return 0;
		*data++ = 1;								// run of 1
		*data   = last;
		count  += 2;
done:

	// Fill header word and return

	return count;
}
